"""Setup"""
import os
import typing

import tornado.web

from tornado_swagger._builders import generate_doc_from_endpoints
from tornado_swagger._handlers import SwaggerSpecHandler, SwaggerUiHandler
from tornado_swagger.const import API_SWAGGER_2

STATIC_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "swagger_ui"))


def export_swagger(
    routes: typing.List[tornado.web.URLSpec],
    *,
    api_base_url: str = "/",
    description: str = "Swagger API definition",
    api_version: str = "1.0.0",
    title: str = "Swagger API",
    contact: str = "",
    schemes: list = None,
    security_definitions: dict = None,
    security: list = None,
    api_definition_version: str = API_SWAGGER_2
):
    """Export swagger schema as dict"""
    return generate_doc_from_endpoints(
        routes,
        api_base_url=api_base_url,
        description=description,
        api_version=api_version,
        title=title,
        contact=contact,
        schemes=schemes,
        security_definitions=security_definitions,
        security=security,
        api_definition_version=api_definition_version,
    )


def setup_swagger(
    routes: typing.List[tornado.web.URLSpec],
    *,
    swagger_url: str = "/api/doc",
    api_base_url: str = "/",
    description: str = "Swagger API definition",
    api_version: str = "1.0.0",
    title: str = "Swagger API",
    contact: str = "",
    schemes: list = None,
    security_definitions: dict = None,
    security: list = None,
    display_models: bool = True,
    api_definition_version: str = API_SWAGGER_2
):
    """Inject swagger ui to application routes"""
    swagger_schema = generate_doc_from_endpoints(
        routes,
        api_base_url=api_base_url,
        description=description,
        api_version=api_version,
        title=title,
        contact=contact,
        schemes=schemes,
        security_definitions=security_definitions,
        security=security,
        api_definition_version=api_definition_version,
    )

    _swagger_ui_url = "/{}".format(swagger_url) if not swagger_url.startswith("/") else swagger_url
    _base_swagger_ui_url = _swagger_ui_url.rstrip("/")
    _swagger_spec_url = "{}/swagger.json".format(_swagger_ui_url)

    routes[:0] = [
        tornado.web.url(_swagger_ui_url, SwaggerUiHandler),
        tornado.web.url("{}/".format(_base_swagger_ui_url), SwaggerUiHandler),
        tornado.web.url(_swagger_spec_url, SwaggerSpecHandler),
    ]

    SwaggerSpecHandler.SWAGGER_SPEC = swagger_schema

    with open(os.path.join(STATIC_PATH, "ui.html"), "r") as f:
        SwaggerUiHandler.SWAGGER_HOME_TEMPLATE = (
            f.read().replace("{{ SWAGGER_URL }}", _swagger_spec_url).replace("{{ DISPLAY_MODELS }}", str(-1 if not display_models else 1))
        )



# pylint: disable=R0401,C0415
import abc
import collections
import inspect
import os
import re
import typing

import tornado.web
import yaml

from tornado_swagger.const import API_OPENAPI_3, API_SWAGGER_2

SWAGGER_TEMPLATE = os.path.abspath(os.path.join(os.path.dirname(__file__), "templates", "swagger.yaml"))
SWAGGER_DOC_SEPARATOR = "---"


def _extract_swagger_definition(endpoint_doc):
    """Extract swagger definition after SWAGGER_DOC_SEPARATOR"""
    endpoint_doc = endpoint_doc.splitlines()

    for i, doc_line in enumerate(endpoint_doc):
        if SWAGGER_DOC_SEPARATOR in doc_line:
            end_point_swagger_start = i + 1
            endpoint_doc = endpoint_doc[end_point_swagger_start:]
            break
    return "\n".join(endpoint_doc)


def build_swagger_docs(endpoint_doc):
    """Build swagger doc based on endpoint docstring"""
    endpoint_doc = _extract_swagger_definition(endpoint_doc)

    # Build JSON YAML Obj
    try:
        endpoint_doc = endpoint_doc.replace("\t", "    ")  # fix windows tabs bug
        end_point_swagger_doc = yaml.safe_load(endpoint_doc)
        if not isinstance(end_point_swagger_doc, dict):
            raise yaml.YAMLError()
        return end_point_swagger_doc
    except yaml.YAMLError:
        return {
            "description": "Swagger document could not be loaded from docstring",
            "tags": ["Invalid Swagger"],
        }


def _try_extract_doc(func):
    """Extract docstring from origin function removing decorators"""
    return inspect.unwrap(func).__doc__


def _build_doc_from_func_doc(handler):
    out = {}

    for method in handler.SUPPORTED_METHODS:
        method = method.lower()
        doc = _try_extract_doc(getattr(handler, method))

        if doc is not None and "---" in doc:
            out.update({method: build_swagger_docs(doc)})

    return out


def _try_extract_args(method_handler):
    """Extract method args from origin function removing decorators"""
    return inspect.getfullargspec(inspect.unwrap(method_handler)).args[1:]


def _extract_parameters_names(handler, parameters_count, method):
    """Extract parameters names from handler"""
    if parameters_count == 0:
        return []

    parameters = ["{?}" for _ in range(parameters_count)]

    method_handler = getattr(handler, method.lower())
    args = _try_extract_args(method_handler)

    for i, arg in enumerate(args):
        if set(arg) != {"_"} and i < len(parameters):
            parameters[i] = arg

    return parameters


def _format_handler_path(route, method):
    brackets_regex = re.compile(r"\(.*?\)")
    parameters = _extract_parameters_names(route.target, route.regex.groups, method)
    route_pattern = route.regex.pattern

    for i, entity in enumerate(brackets_regex.findall(route_pattern)):
        route_pattern = route_pattern.replace(entity, "{%s}" % parameters[i], 1)

    return route_pattern[:-1]


def nesteddict2yaml(d, indent=10, result=""):
    for key, value in d.items():
        result += " " * indent + str(key) + ":"
        if isinstance(value, dict):
            result = nesteddict2yaml(value, indent + 2, result + "\n")
        else:
            result += " " + str(value) + "\n"
    return result


def _clean_description(description):
    """Remove empty space from description begin"""
    _start_desc = 0
    for i, word in enumerate(description):
        if word != "\n":
            _start_desc = i
            break
    return "    ".join(description[_start_desc:].splitlines())


def _extract_paths(routes):
    paths = collections.defaultdict(dict)

    for route in routes:
        for method_name, method_description in _build_doc_from_func_doc(route.target).items():
            paths[_format_handler_path(route, method_name)].update({method_name: method_description})

    return paths


class BaseDocBuilder(abc.ABC):
    """Doc builder"""

    @property
    @abc.abstractmethod
    def schema(self):
        """Supported Schema"""

    @abc.abstractmethod
    def generate_doc(
        self,
        routes: typing.List[tornado.web.URLSpec],
        *,
        api_base_url,
        description,
        api_version,
        title,
        contact,
        schemes,
        security_definitions,
        security,
        models,
        parameters
    ):
        """Generate docs"""


class Swagger2DocBuilder(BaseDocBuilder):
    """Swagger2.0 schema builder"""

    @property
    def schema(self):
        """Supported Schema"""
        return API_SWAGGER_2

    def generate_doc(
        self,
        routes: typing.List[tornado.web.URLSpec],
        *,
        api_base_url,
        description,
        api_version,
        title,
        contact,
        schemes,
        security_definitions,
        security,
        models,
        parameters
    ):
        """Generate docs"""
        swagger_spec = {
            "swagger": "2.0",
            "info": {
                "title": title,
                "description": _clean_description(description),
                "version": api_version,
            },
            "basePath": api_base_url,
            "schemes": schemes,
            "definitions": models,
            "parameters": parameters,
            "paths": _extract_paths(routes),
        }
        if contact:
            swagger_spec["info"]["contact"] = {"name": contact}
        if security_definitions:
            swagger_spec["securityDefinitions"] = security_definitions
        if security:
            swagger_spec["security"] = security

        return swagger_spec


class OpenApiDocBuilder(BaseDocBuilder):
    """OpenAPI 3 Schema builder"""

    @property
    def schema(self):
        """Supported Schema"""
        return API_OPENAPI_3

    def generate_doc(
        self,
        routes: typing.List[tornado.web.URLSpec],
        *,
        api_base_url,
        description,
        api_version,
        title,
        contact,
        schemes,
        security_definitions,
        security,
        models,
        parameters
    ):
        """Generate docs"""
        swagger_spec = {
            "openapi": "3.0.3",
            "info": {
                "title": title,
                "description": _clean_description(description),
                "version": api_version,
            },
            "basePath": api_base_url,
            "schemes": schemes,
            "components": {
                "schemas": models,
                "parameters": parameters,
            },
            "paths": _extract_paths(routes),
        }

        if contact:
            swagger_spec["info"]["contact"] = {"name": contact}
        if security_definitions:
            swagger_spec["securityDefinitions"] = security_definitions
        if security:
            swagger_spec["security"] = security

        return swagger_spec


doc_builders = {b.schema: b for b in [Swagger2DocBuilder(), OpenApiDocBuilder()]}


def generate_doc_from_endpoints(
    routes: typing.List[tornado.web.URLSpec],
    *,
    api_base_url,
    description,
    api_version,
    title,
    contact,
    schemes,
    security_definitions,
    security,
    api_definition_version
):
    """Generate doc based on routes"""
    from tornado_swagger.model import export_swagger_models
    from tornado_swagger.parameter import export_swagger_parameters

    if api_definition_version not in doc_builders:
        raise ValueError("Unknown api_definition_version = " + api_definition_version)

    return doc_builders[api_definition_version].generate_doc(
        routes,
        api_base_url=api_base_url,
        description=description,
        api_version=api_version,
        title=title,
        contact=contact,
        schemes=schemes,
        security_definitions=security_definitions,
        security=security,
        models=export_swagger_models(),
        parameters=export_swagger_parameters(),
    )



import tornado.web


class TornadoBaseHandler(tornado.web.RequestHandler):
    def data_received(self, chunk: bytes):
        pass


class SwaggerUiHandler(TornadoBaseHandler):
    SWAGGER_HOME_TEMPLATE = ""

    def get(self):
        self.write(self.SWAGGER_HOME_TEMPLATE)


class SwaggerSpecHandler(TornadoBaseHandler):
    SWAGGER_SPEC = ""

    def get(self):
        self.write(self.SWAGGER_SPEC)



from tornado_swagger._builders import build_swagger_docs


class _SwaggerModelsStore:
    """Singleton with models definitions"""

    definitions = {}


def _save_model_doc(model):
    """Save model docstring to _SwaggerModelsStore"""
    doc = model.__doc__

    if doc is not None and "---" in doc:
        _SwaggerModelsStore.definitions[model.__name__] = build_swagger_docs(doc)


def export_swagger_models():
    """Get swagger models definition"""
    return _SwaggerModelsStore.definitions


def register_swagger_model(model):
    """Register model definition in swagger"""
    _save_model_doc(model)
    return model



from os.path import dirname
from os.path import join

from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test as TestCommand

from tornado_swagger import __version__

PACKAGE_NAME = "tornado-swagger"
DESCRIPTION = "Swagger API Documentation builder for tornado server"
HOME_URL = "https://github.com/mrk-andreev/tornado-swagger"
DOWNLOAD_URL = "https://pypi.org/project/tornado-swagger/#files"
MAINTAINER = "Mark Andreev"
MAINTAINER_EMAIL = "mark.andreev@gmail.com"
LICENSE = "MIT License"
CLASSIFIERS = [
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

with open(join(dirname(__file__), "requirements.txt")) as f:
    PACKAGES_REQUIRED = f.read().splitlines()

with open(join(dirname(__file__), "README.rst")) as f:
    LONG_DESCRIPTION = f.read()


class PyTest(TestCommand):
    user_options = []

    def run(self):
        import subprocess
        import sys

        errno = subprocess.call([sys.executable, "-m", "pytest", "tests"])
        raise SystemExit(errno)


def setup_package():
    setup(
        name=PACKAGE_NAME,
        version=__version__,
        install_requires=PACKAGES_REQUIRED,
        url=HOME_URL,
        download_url=DOWNLOAD_URL,
        license=LICENSE,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        packages=find_packages(),
        include_package_data=True,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        classifiers=CLASSIFIERS,
        tests_require=["pytest"],
        cmdclass=dict(test=PyTest),
    )


if __name__ == "__main__":
    setup_package()



from tornado_swagger._builders import build_swagger_docs


class _SwaggerParameterStore:
    """Singleton with parameter definitions"""

    definitions = {}


def _save_parameter_doc(model):
    """Save model docstring to _SwaggerParameterStore"""
    doc = model.__doc__

    if doc is not None and "---" in doc:
        _SwaggerParameterStore.definitions[model.__name__] = build_swagger_docs(doc)


def export_swagger_parameters():
    """Get swagger parameters definition"""
    return _SwaggerParameterStore.definitions


def register_swagger_parameter(model):
    """Register parameter definition in swagger"""
    _save_parameter_doc(model)
    return model
